section \<open>Insertion Sort\<close>

text \<open>Insertion sort as example of "HOL = Functional Programming + Logic".\<close>

theory Insertion_Sort
imports Main
begin

text \<open>Inserting a new element into a (supposedly) sorted list.\<close>

fun ins :: "nat \<Rightarrow> nat list \<Rightarrow> nat list"
where
  "ins x [] = [x]"
| "ins x (y # ys) = (if x \<le> y then x # y # ys else y # ins x ys)"


text \<open>Standard folding operations on lists are part of the library.
  In particular:

    @{term "foldr :: ('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b \<Rightarrow> 'b"}

  Thus, we obtain insertion sort by folding the insertion operation from
  above over a given list.\<close>

definition isort :: "nat list \<Rightarrow> nat list"
  where "isort xs = foldr ins xs []"


text \<open>Main Isabelle/HOL already defines a predicate @{const sorted};
  it indicates whether a given list is sorted (according to
  the standard linear order of the underlying type of elements).

  It is defined inductively by the two rules:

    @{lemma "sorted []" by (rule sorted.Nil)}
    @{lemma "\<forall>y \<in> set xs. x \<le> y \<Longrightarrow> sorted xs \<Longrightarrow> sorted (x # xs)"
        by (rule sorted.Cons)}

  As a derived fact we obtain the equation

    @{lemma "sorted (x # xs) \<longleftrightarrow> sorted xs \<and> (\<forall>y \<in> set xs. x \<le> y)"
        by (rule sorted_Cons)}

  which is available under the name @{thm sorted_Cons}.\<close>


subsection \<open>Exercise: the result of @{const isort} is sorted\<close>

text \<open>
  \<triangleright> Prove @{prop "sorted (isort xs)"} (by identifying and proving
    missing facts).

  Hint: First try induction followed by "auto", then investigate
  the remaining subgoal(s).
\<close>


subsection \<open>Exercise: result is permutation of input\<close>

text \<open>
  Is this really enough for correctness of @{const isort}?

  Consider @{term "isort' xs = []"}.

  Make sure that result is permutation of input.
\<close>

fun count :: "'a \<Rightarrow> 'a list \<Rightarrow> nat"  ("\<sharp>_'(_')")
where
  "\<sharp>x([]) = 0"
| "\<sharp>x(y#ys) = (if x = y then Suc (\<sharp>x(ys)) else \<sharp>x(ys))"

text \<open>
  \<triangleright> Prove that @{const isort} does not change the multiplicity
    of elements according to @{term count}.
\<close>


subsection \<open>Exercise: general types with types classes\<close>

text \<open>
  \<triangleright> Generalize the types of the above definitions as far as possible.
\<close>


subsection \<open>Exercise: alternative definition of @{term sorted}\<close>

text \<open>
  \<triangleright> Define a function
    @{term "is_sorted :: 'a::linorder list \<Rightarrow> bool"}
    and prove that it is equivalent to the inductive
    predicate @{term sorted}.
\<close>

end
